# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from collections import defaultdict
import numpy as np
from tqdm import tqdm
import warnings
from .warnings import  NotIntendedUseWarning


class TreePathFinder():
    """
    Class to calculate the boundaries of a decision tree.
    It retrieves the structure from the decision tree

    """

    def __init__(self, estimator):
        self.estimator = estimator
        self.n_nodes = estimator.tree_.node_count
        self.children_left = estimator.tree_.children_left
        self.children_right = estimator.tree_.children_right
        self.feature = estimator.tree_.feature
        self.threshold = estimator.tree_.threshold
        self.is_leaves = self._find_leaves()
        self.decision_path = self.find_decision_to_leaves()
        self.bin_boundaries = self.find_bin_boundaries()


    def _find_leaves(self):
        # The tree structure can be traversed to compute various properties such
        # as the depth of each node and whether or not it is a leaf.
        n_nodes = self.n_nodes
        children_left = self.children_left
        children_right = self.children_right

        node_depth = np.zeros(shape=n_nodes, dtype=np.int64)
        is_leaves = np.zeros(shape=n_nodes, dtype=bool)
        stack = [(0, -1)]  # seed is the root node id and its parent depth
        while len(stack) > 0:
            node_id, parent_depth = stack.pop()
            node_depth[node_id] = parent_depth + 1

            # If we have a test node
            if (children_left[node_id] != children_right[node_id]):
                stack.append((children_left[node_id], parent_depth + 1))
                stack.append((children_right[node_id], parent_depth + 1))
            else:
                is_leaves[node_id] = True

        return is_leaves

    def find_parent(self, leaf_id):

        in_left = np.where(self.children_left == leaf_id)[0]
        in_right = np.where(self.children_right == leaf_id)[0]
        is_inleft = len(in_left) > 0
        is_inright = len(in_right) > 0

        if is_inleft & is_inright:
            raise ValueError(f"leaf with id {leaf_id} not found in tree")
        elif is_inleft:
            parent = in_left[0]
            operator = '<='
        elif is_inright:
            parent = in_right[0]
            operator = '>'
        else:
            parent = 0
            operator = 'None'

        threshold = self.threshold[parent]
        feature = self.feature[parent]

        return parent, threshold, operator, feature


    def find_decision_to_leaves(self):

        leaves_ids = np.where(self.is_leaves)[0]

        decision_path = defaultdict(list)
        for leaf_id in tqdm(leaves_ids):
            parent_id = -1
            node_id = leaf_id
            while parent_id != 0:
                path_step = self.find_parent(node_id)
                parent_id = path_step[0]
                decision_path[leaf_id].append(path_step)
                node_id = parent_id

        return decision_path

    def find_bin_boundaries(self):

        out_dict = dict()
        for leaf_id in self.decision_path.keys():
            one_leaf_decisions = self.decision_path[leaf_id]

            if '<=' not in [oper[2] for oper in one_leaf_decisions]:
                max_val = np.inf
            else:
                max_val = min([oper[1] for oper in one_leaf_decisions if oper[2] == '<='])

            if '>' not in [oper[2] for oper in one_leaf_decisions]:
                min_val = -np.inf
            else:
                min_val = max([oper[1] for oper in one_leaf_decisions if oper[2] == '>'])


            out_dict[leaf_id] = {
                'min': min_val,
                'max': max_val,
            }

        return out_dict


    def get_boundaries(self):

        # check how many features are there. There is always a unique negative value in the array of features
        # that corresponds to the index of the leaves.
        # Hence the total number of features in the tree is the length of the array -1
        n_features = len(np.unique(self.feature))-1

        if n_features>1:
            warning = f"This functionality is intended for trees fitted on 1 feature. The current tree is fitted " \
                f"with {n_features} features"
            warnings.warn(NotIntendedUseWarning(warning))

        return self.bin_boundaries


# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import numpy as np


def sample_data(X, y, sampling_type, sampling_fraction, dataset_name='dataset'):
    check_sampling_input(sampling_type, sampling_fraction, dataset_name)

    if sampling_type is None:
        return X, y

    number_of_samples = np.ceil(sampling_fraction * X.shape[0]).astype(int)
    array_index = list(range(X.shape[0]))

    if sampling_type == 'bootstrap':
        rows_indexes = np.random.choice(array_index, number_of_samples, replace=True)
    else:
        if sampling_fraction == 1 or number_of_samples == X.shape[0]:
            return X,y
        else:
            rows_indexes = np.random.choice(array_index, number_of_samples, replace=True)
    return X[rows_indexes], y[rows_indexes]


def check_sampling_input(sampling_type, fraction, dataset_name):
    if sampling_type is not None:
        if sampling_type == 'bootstrap':
            if fraction <= 0:
                raise(ValueError(f'For bootstrapping {dataset_name} fraction needs to be above 0'))
        elif sampling_type == 'subsample':
            if  fraction <= 0 or fraction >= 1:
                raise(ValueError(f'For bootstrapping {dataset_name} fraction needs to be be above 0 and below 1'))
        else:
            raise(ValueError('This sampling method is not implemented'))


# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


def class_name_from_object(obj):
    return obj.__class__.__name__

def assure_list_of_strings(variable, variable_name):
    if isinstance(variable, list):
        return variable
    elif isinstance(variable, str):
        return [variable]
    else:
        raise(ValueError('{} needs to be either a string or list of strings.').format(variable_name))


def assure_list_values_allowed(variable, variable_name, allowed_values):
    for value in variable:
        if value not in allowed_values:
            raise(ValueError('Value {} in variable {} is not allowed').format(value, variable_name))



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import pandas as pd
import numpy as np
from sklearn.cluster import AgglomerativeClustering
from sklearn.tree import DecisionTreeClassifier
from sklearn.utils.validation import check_is_fitted

from probatus.utils import assure_numpy_array, TreePathFinder, ApproximationWarning

import warnings
from abc import ABC, abstractmethod

class Bucketer(ABC):

    def __repr__(self):
        repr_ = f"{self.__class__.__name__}\n\tbincount: {self.bin_count}"
        if hasattr(self, "boundaries_"):
            repr_ += f"\nResults:\n\tcounts: {self.counts_}\n\tboundaries: {self.boundaries_}"
        return repr_

    @abstractmethod
    def fit(self):
        pass

    @property
    def boundaries(self):
        warnings.warn("The 'boundaries' attribute is deprecated, use 'boundaries_' instead. The underscore suffix signals this is a fitted attribute.",
                       DeprecationWarning)
        check_is_fitted(self)
        return self.boundaries_

    @property
    def counts(self):
        warnings.warn("The 'counts' attribute is deprecated, use 'counts_' instead. The underscore suffix signals this is a fitted attribute.", 
            DeprecationWarning)
        check_is_fitted(self)
        return self.counts_

    def compute(self, X, y=None):
        """
        Applies fitted bucketing algorithm on input data and counts number of samples per bin

        Args:
            X: (np.array) data to be bucketed
            y: (np.array) ignored, for sklearn compatibility

        Returns: counts of the elements in X using the bucketing that was obtained by fitting the Bucketer instance

        """
        check_is_fitted(self)
        
        # np.digitize returns the indices of the bins to which each value in input array belongs
        # the smallest value of the `boundaries` attribute equals the lowest value in the set the instance was
        # fitted on, to prevent the smallest value of x_new to be in his own bucket, we ignore the first boundary
        # value
        digitize_result = np.digitize(X, self.boundaries_[1:], right=True)
        result = pd.DataFrame({'bucket': digitize_result}).groupby('bucket')['bucket'].count()
        # reindex the dataframe such that also empty buckets are included in the result
        result = result.reindex(np.arange(self.bin_count), fill_value=0)
        return result.values

    def fit_compute(self, X, y = None):
        """
        Apply bucketing to new data and return number of samples per bin

        Args:
            X: (np.array) data to be bucketed
            y: (np.array) One dimensional array, used if the target is needed for the bucketing. By default is set to
            None

        Returns: counts of the elements in x_new using the bucketing that was obtained by fitting the Bucketer instance

        """
        self.fit(X, y)
        return self.compute(X, y)


class SimpleBucketer(Bucketer):
    """Create equally spaced bins using numpy.histogram function

    Example:
    ```python
    x = [1, 2, 1]
    bins = 3
    myBucketer = SimpleBucketer(bin_count=bins)
    myBucketer.fit(x)
    ```

    myBucketer.counts gives the number of elements per bucket
    myBucketer.boundaries gives the boundaries of the buckets
    """
    def __init__(self, bin_count):
        self.bin_count = bin_count

    @staticmethod
    def simple_bins(x, bin_count):
        counts, boundaries = np.histogram(x, bins=bin_count)
        return counts, boundaries

    def fit(self, x, y=None):
        """
        Fit bucketing on x

        Args:
            x: (np.array) Input array on which the boundaries of bins are fitted
            y: (np.array) ignored. For sklearn-compatibility

        Returns: fitted bucketer object
        """
        self.counts_, self.boundaries_ = self.simple_bins(x, self.bin_count)
        return self


class AgglomerativeBucketer(Bucketer):
    """Create binning by applying the Scikit-learn implementation of Agglomerative Clustering

    Usage:
    ```python
    x = [1, 2, 1]
    bins = 3
    myBucketer = AgglomerativeBucketer(bin_count=bins)
    myBucketer.fit(x)
    ```

    myBucketer.counts gives the number of elements per bucket
    myBucketer.boundaries gives the boundaries of the buckets
    """
    def __init__(self, bin_count):
        self.bin_count = bin_count

    @staticmethod
    def agglomerative_clustering_binning(x, bin_count):
        clustering = AgglomerativeClustering(n_clusters=bin_count).fit(np.asarray(x).reshape(-1, 1))
        df = pd.DataFrame({'x': x, 'label': clustering.labels_}).sort_values(by='x')
        cluster_minimum_values = df.groupby('label')['x'].min().sort_values().tolist()
        cluster_maximum_values = df.groupby('label')['x'].max().sort_values().tolist()
        # take the mean of the upper boundary of a cluster and the lower boundary of the next cluster
        boundaries = [np.mean([cluster_minimum_values[i + 1], cluster_maximum_values[i]]) for i in
                      range(len(cluster_minimum_values) - 1)]
        # add the lower boundary of the lowest cluster and the upper boundary of the highest cluster
        boundaries = [cluster_minimum_values[0]] + boundaries + [cluster_maximum_values[-1]]
        counts = df.groupby('label')['label'].count().values
        return counts, boundaries

    def fit(self, x, y=None):
        """
        Fit bucketing on x

        Args:
            x: (np.array) Input array on which the boundaries of bins are fitted
            y: (np.array) ignored. For sklearn-compatibility

        Returns: fitted bucketer object
        """
        self.counts_, self.boundaries_ = self.agglomerative_clustering_binning(x, self.bin_count)
        return self


class QuantileBucketer(Bucketer):
    """Create bins with equal number of elements

    Usage:
    ```python
    x = [1, 2, 1]
    bins = 3
    myBucketer = QuantileBucketer(bin_count=bins)
    myBucketer.fit(x)
    ```

    myBucketer.counts gives the number of elements per bucket
    myBucketer.boundaries gives the boundaries of the buckets
    """
    def __init__(self, bin_count):
        self.bin_count = bin_count

    @staticmethod
    def quantile_bins(x, bin_count, inf_edges=False):

        try:
            out, boundaries = pd.qcut(x, q=bin_count, retbins=True, duplicates='raise')
        except ValueError:
            # If there are too many duplicate values (assume a lot of filled missings)
            # this crashes - the exception drops them.
            # This means that it will return approximate quantile bins
            out, boundaries = pd.qcut(x, q=bin_count, retbins=True, duplicates='drop')
            warnings.warn(ApproximationWarning("Approximated quantiles - too many unique values" ))
        df = pd.DataFrame({'x': x})
        df['label'] = out
        counts = df.groupby('label').count().values.flatten()
        if inf_edges:
            boundaries[0] = -np.inf
            boundaries[-1] = np.inf
        return counts, boundaries

    def fit(self, x, y=None):
        """
        Fit bucketing on x

        Args:
            x: (np.array) Input array on which the boundaries of bins are fitted
            y: (np.array) ignored. For sklearn-compatibility

        Returns: fitted bucketer object
        """
        self.counts_, self.boundaries_ = self.quantile_bins(x, self.bin_count)
        return self


class TreeBucketer(Bucketer):
    """Class for bucketing using Decision Trees.
    It returns the optimal buckets found by a one-dimensional Decision Tree relative to a binary target.

    Useful if the buckets be defined such that there is a substantial difference between the buckets in
    the distribution of the target.
    
    Usage:
    ```python
    x = [1, 2, 2, 5 ,3]
    y = [0, 0 ,1 ,1 ,1]
    bins = 3
    myBucketer = TreeBucketer(inf_edges=True,max_depth=2,min_impurity_decrease=0.001)
    myBucketer.fit(x,y)
    ```

    myBucketer.counts gives the number of elements per bucket
    myBucketer.boundaries gives the boundaries of the buckets

    Args:
        inf_edges (boolean): Flag to keep the lower and upper boundary as infinite (if set to True).
        If false, the edges will be set to the minimum and maximum value of the fitted

        tree (sklearn.tree.DecisionTreeClassifier): decision tree object defined by the user. By default is None, and
        it will be constructed using tkhe provided **kwargs

        **tree_kwargs: kwargs related to the decision tree.
            For and extensive list of parameteres, please check the sklearn Decision Tree Classifier documentation
            https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeClassifier.html

            The most relevant parameteres useful for the bucketing, are listed below:


                - criterion : {"gini", "entropy"}, default="gini"
                    The function to measure the quality of a split. Supported criteria are
                    "gini" for the Gini impurity and "entropy" for the information gain.


                - max_depth : int, default=None
                    Defines the maximum theoretical number of bins (2^max_depth)

                    The maximum depth of the tree. If None, then nodes are expanded until
                    all leaves are pure or until all leaves contain less than
                    min_samples_split samples.



                - min_samples_leaf : int or float, default=1
                    Defines the minimum number of entries in each bucket.

                    The minimum number of samples required to be at a leaf node.
                    A split point at any depth will only be considered if it leaves at
                    least ``min_samples_leaf`` training samples in each of the left and
                    right branches.  This may have the effect of smoothing the model,
                    especially in regression.

                    - If int, then consider `min_samples_leaf` as the minimum number.
                    - If float, then `min_samples_leaf` is a fraction and
                      `ceil(min_samples_leaf * n_samples)` are the minimum
                      number of samples for each node.

                    .. versionchanged:: 0.18
                       Added float values for fractions.



                min_impurity_decrease : float, default=0.0
                    Controls the way the TreeBucketer splits.
                    When the criterion is set to 'entropy', the best results tend to
                    be achieved in the range [0.0001 - 0.01]

                    A node will be split if this split induces a decrease of the impurity
                    greater than or equal to this value.

                    The weighted impurity decrease equation is the following::

                        N_t / N * (impurity - N_t_R / N_t * right_impurity
                                            - N_t_L / N_t * left_impurity)

                    where ``N`` is the total number of samples, ``N_t`` is the number of
                    samples at the current node, ``N_t_L`` is the number of samples in the
                    left child, and ``N_t_R`` is the number of samples in the right child.

                    ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
                    if ``sample_weight`` is passed.

                    .. versionadded:: 0.19

    """

    def __init__(self, inf_edges = False, tree = None, **tree_kwargs):
        self.bin_count = -1
        self.inf_edges = inf_edges
        if tree is None:
            self.tree = DecisionTreeClassifier(**tree_kwargs)
        else:
            self.tree = tree


    @staticmethod
    def tree_bins(x, y, inf_edges, tree):

        X_in = assure_numpy_array(x).reshape(-1, 1)
        y_in = assure_numpy_array(y).reshape(-1, 1)
        tree.fit(X_in,y_in)

        if tree.min_samples_leaf>=X_in.shape[0]:
            error_msg = (
                "Cannot Fit decision tree. min_samples_leaf must be < than the length of x.m" +
                f"Currently min_samples_leaf {tree.min_samples_leaf} " +
                f"and the length of X is {X_in.shape[0]}"
            )
            raise ValueError(error_msg)

        leaves = tree.apply(X_in)
        index, counts = np.unique(leaves, return_counts=True)

        bin_count = len(index)

        tpf = TreePathFinder(tree)
        boundaries = [bound['min'] for bound in tpf.get_boundaries().values()]
        boundaries += [tpf.get_boundaries()[leaves[-1]]['max']]

        if not inf_edges:
            boundaries[0] = np.min(X_in)
            boundaries[-1] = np.max(X_in)

        return counts, boundaries, bin_count, tree

    def fit(self, X, y):
        """
        Fit bucketing on x

        Args:
            x: (np.array) Input array on which the boundaries of bins are fitted
            y: (np.array) optional, One dimensional array with the target.

        Returns: fitted bucketer object
        """
        self.counts_, self.boundaries_, self.bin_count, self.tree = self.tree_bins(X,y, self.inf_edges, self.tree)
        return self


# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import shap
import pandas as pd
import numpy as np
import warnings

def shap_calc(model, X, approximate=False, return_explainer=False, suppress_warnings=False, **shap_kwargs):
    """
    Helper function to calculate the shapley values for a given model.
    Supported models for the moment are RandomForestClassifiers and XGBClassifiers
    In case the shapley values have
    Args:
        model: pretrained model (Random Forest of XGBoost at the moment)
        X (pd.DataFrame or np.ndarray): features set
        approximate (boolean):, if True uses shap approximations - less accurate, but very fast
        return_explainer (boolean): if True, returns a a tuple (shap_values, explainer).
        suppress_warnings (boolean): If True, warnings from SHAP will be suppressed.
        **shap_kwargs: kwargs of the shap.TreeExplainer

    Returns: (np.ndarray or tuple(np.ndarray, shap.TreeExplainer)) shapley_values for the model.

    """
    # Suppress warnings regarding XGboost and Lightgbm models.
    with warnings.catch_warnings():
        if suppress_warnings:
            warnings.simplefilter("ignore")

        explainer = shap.TreeExplainer(model, **shap_kwargs)
        # Calculate Shap values
        shap_values = explainer.shap_values(X, approximate=approximate)

        if isinstance(shap_values, list) and len(shap_values)==2:
            warnings.warn('Shap values are related to the output probabilities of class 1 for this model, instead of log odds.')
            shap_values = shap_values[1]

    if return_explainer:
        return shap_values, explainer
    return shap_values


def shap_to_df(model, X, precalc_shap=None, **kwargs):
    """
    Calculates the shap values and return the pandas DataFrame with the columns and the index of the original

    Args:
        model: pretrained model (Random Forest of XGBoost at the moment)
        X (pd.DataFrame or np.ndarray): features set
        precalc_shap (np.array): Precalculated SHAP values. If None, they are computed.
        **kwargs: for the function shap_calc

    Returns:

    """
    if precalc_shap is not None:
        shap_values = precalc_shap
    else:
        shap_values = shap_calc(model, X, **kwargs)
    if isinstance(X, pd.DataFrame):
        return pd.DataFrame(shap_values, columns=X.columns, index=X.index)

    elif isinstance(X, np.ndarray) and len(X.shape) == 2:
        return pd.DataFrame(shap_values, columns=[f"col_{ix}" for ix in range(X.shape[1])])

    else:
        raise NotImplementedError("X must be a dataframe or a 2d array")


def calculate_shap_importance(shap_values, columns):
    """
    Returns the average shapley value for each column of the dataframe, as well as the average absolute shap value.

    Args:
        shap_values (np.array): Shap values.
        columns (list of str): Feature names.

    Returns:
        (pd.DataFrame): Mean absolute shap values and Mean shap values of features.

    """

    # Find average shap importance for neg and pos class
    shap_abs_mean = np.mean(np.abs(shap_values), axis=0)
    shap_mean = np.mean(shap_values, axis=0)

    # Prepare importance values in a handy df
    importance_df = pd.DataFrame({
        'mean_abs_shap_value':shap_abs_mean.tolist(),
        'mean_shap_value': shap_mean.tolist()},
        index=columns)

    # Set the correct column types
    importance_df['mean_abs_shap_value'] = importance_df['mean_abs_shap_value'].astype(float)
    importance_df['mean_shap_value'] = importance_df['mean_shap_value'].astype(float)

    importance_df = importance_df.sort_values('mean_abs_shap_value', ascending=False)

    return importance_df



import numpy as np
import pandas as pd

import pytest

from probatus.utils import assure_numpy_array, assure_pandas_df, check_1d, DimensionalityError, check_numeric_dtypes


def test_assure_numpy_array_list():
    x = [1, 2, 3]
    x_array = assure_numpy_array(x)
    assert isinstance(x_array, np.ndarray)
    np.testing.assert_array_equal(x_array, np.array(x))
    x = [[1, 2], [3, 4]]
    x_array = assure_numpy_array(x)
    np.testing.assert_array_equal(x_array, np.array([[1, 2], [3, 4]]))
    with pytest.raises(DimensionalityError):
        assert assure_numpy_array(x, assure_1d=True)


def test_assure_numpy_array_array():
    x = np.array([1, 2, 3])
    x_array = assure_numpy_array(x)
    assert isinstance(x_array, np.ndarray)
    np.testing.assert_array_equal(x_array, x)
    x = np.array([[1, 2], [3, 4]])
    x_array = assure_numpy_array(x)
    np.testing.assert_array_equal(x_array, x)
    with pytest.raises(DimensionalityError):
        assert assure_numpy_array(x, assure_1d=True)


def test_assure_numpy_array_dataframe():
    x = pd.DataFrame({'x': [1, 2, 3]})
    x_array = assure_numpy_array(x)
    assert isinstance(x_array, np.ndarray)
    np.testing.assert_array_equal(x_array, np.array([1, 2, 3]))
    x = pd.DataFrame({'x': [1, 2, 3], 'y': [1, 2, 3]})
    x_array = assure_numpy_array(x)
    np.testing.assert_array_equal(x_array, np.array([[1, 1], [2, 2], [3, 3]]))
    with pytest.raises(DimensionalityError):
        assert assure_numpy_array(x, assure_1d=True)


def test_assure_numpy_array_series():
    x = pd.Series([1, 2, 3])
    x_array = assure_numpy_array(x)
    assert isinstance(x_array, np.ndarray)
    np.testing.assert_array_equal(x_array, np.array([1, 2, 3]))


def test_check_1d_list():
    x = [1, 2, 3]
    assert check_1d(x)
    y = [[1, 2], [1, 2, 3]]
    with pytest.raises(DimensionalityError):
        assert check_1d(y)
    y = [1, [1, 2, 3]]
    with pytest.raises(DimensionalityError):
        assert check_1d(y)


def test_check_1d_array():
    x = np.array([1, 2, 3])
    assert check_1d(x)
    y = np.array([[1, 2], [1, 2, 3]])
    with pytest.raises(DimensionalityError):
        assert check_1d(y)
    y = np.array([0, [1, 2, 3]])
    with pytest.raises(DimensionalityError):
        assert check_1d(y)


def test_check_1d_dataframe():
    x = pd.DataFrame({'x': [1, 2, 3]})
    assert check_1d(x)
    y = pd.DataFrame({'x': [1, 2, 3], 'y': [1, 2, 3]})
    with pytest.raises(DimensionalityError):
        assert check_1d(y)
    y = pd.DataFrame({'x': [1, 2, 3, [4, 5]]})
    with pytest.raises(DimensionalityError):
        assert check_1d(y)


def test_check_1d_series():
    x = pd.Series([1, 2, 3])
    assert check_1d(x)
    y = pd.Series([1, [2, 3]])
    with pytest.raises(DimensionalityError):
        assert check_1d(y)
    y = pd.Series([[1], [2, 3]])
    with pytest.raises(DimensionalityError):
        assert check_1d(y)

@pytest.fixture(scope="function")
def expected_df_2d():
    return pd.DataFrame({0: [1, 2], 1: [2, 3], 2: [3, 4]})

@pytest.fixture(scope="function")
def expected_df():
    return pd.DataFrame({0: [1, 2, 3]})

def test_assure_pandas_df_list(expected_df):
    x = [1, 2, 3]
    x_df = assure_pandas_df(x)
    assert x_df.equals(expected_df)

def test_assure_pandas_df_list_of_lists(expected_df_2d):
    x = [[1, 2, 3], [2, 3, 4]]
    x_df = assure_pandas_df(x)
    assert x_df.equals(expected_df_2d)
    
def test_assure_pandas_df_series(expected_df):
    x = pd.Series([1, 2, 3])
    x_df = assure_pandas_df(x)
    assert x_df.equals(expected_df)
    
def test_assure_pandas_df_array(expected_df, expected_df_2d):
    x = np.array([[1, 2, 3], [2, 3, 4]])
    x_df = assure_pandas_df(x)
    assert x_df.equals(expected_df_2d)
    
    x = np.array([1, 2, 3])
    x_df = assure_pandas_df(x)
    assert x_df.equals(expected_df)
    
def test_assure_pandas_df_df(expected_df_2d):
    x = pd.DataFrame([[1, 2, 3], [2, 3, 4]])
    x_df = assure_pandas_df(x)
    assert x_df.equals(expected_df_2d)
    
def test_assure_pandas_df_types():
    with pytest.raises(TypeError):
        assure_pandas_df("Test")
    with pytest.raises(TypeError):
        assure_pandas_df(5)

def test_check_numeric_dtype_list():
    with pytest.raises(TypeError):
        check_numeric_dtypes(['not numeric', 7, 1.0, True])
    check_numeric_dtypes([1, 2, 3])
    check_numeric_dtypes([1.0, 2.0, 3.0])
    check_numeric_dtypes([False, True, False])
    check_numeric_dtypes([1, True, 7.0])
    
        


from unittest.mock import Mock


# These are shell classes that define the methods of the models that we use. Each of the functions that we use needs
# To be defined inside these shell classes. Then when we want to write a specific test you need to simply mock.patch
# the desired functionality. You can also set the return_value to the patched method.

class MockClusterer(Mock):
    def __init__(self):
        pass

    def fit(self):
        pass

    def predict(self):
        pass

    def fit_predict(self):
        pass

class MockModel(Mock):
    def __init__(self, **kwargs):
        pass

    def fit(self):
        pass

    def predict(self):
        pass

    def predict_proba(self):
        pass





from sklearn.tree import DecisionTreeClassifier
import pytest
import numpy as np
import pandas as pd
from probatus.feature_elimination import ShapRFECV
from sklearn.model_selection import RandomizedSearchCV
from sklearn.metrics import get_scorer

@pytest.fixture(scope='function')
def X():
    return pd.DataFrame({'col_1': [1, 1, 1, 1, 1, 1, 1, 0],
                         'col_2': [0, 0, 0, 0, 0, 0, 0, 1],
                         'col_3': [1, 0, 1, 0, 1, 0, 1, 0]}, index=[1, 2, 3, 4, 5, 6, 7, 8])


@pytest.fixture(scope='function')
def y():
    return pd.Series([1, 0, 1, 0, 1, 0, 1, 0], index=[1, 2, 3, 4, 5, 6, 7, 8])


def test_shap_rfe_randomized_search(X, y):

    clf = DecisionTreeClassifier(max_depth=1)
    param_grid = {
        'criterion': ['gini', 'entropy'],
        'min_samples_split': [1, 2]
    }
    search = RandomizedSearchCV(clf, param_grid, cv=2)

    shap_elimination = ShapRFECV(search, step=0.8, cv=2, scoring='roc_auc', n_jobs=4)

    report = shap_elimination.fit_compute(X, y)

    assert shap_elimination.fitted == True
    shap_elimination._check_if_fitted

    assert report.shape[0] == 2
    assert shap_elimination.get_reduced_features_set(1) == ['col_3']

    ax1 = shap_elimination.plot(show=False)


def test_shap_rfe(X, y):

    clf = DecisionTreeClassifier(max_depth=1)

    shap_elimination = ShapRFECV(clf, random_state=1, step=1, cv=2, scoring='roc_auc', n_jobs=4)

    shap_elimination.fit(X, y)

    assert shap_elimination.fitted == True
    shap_elimination._check_if_fitted

    report = shap_elimination.compute()

    assert report.shape[0] == 3
    assert shap_elimination.get_reduced_features_set(1) == ['col_3']

    ax1 = shap_elimination.plot(show=False)


def test_calculate_number_of_features_to_remove():
    assert 3 == ShapRFECV._calculate_number_of_features_to_remove(current_num_of_features=10,
                                                                  num_features_to_remove=3,
                                                                  min_num_features_to_keep=5)
    assert 3 == ShapRFECV._calculate_number_of_features_to_remove(current_num_of_features=8,
                                                                  num_features_to_remove=5,
                                                                  min_num_features_to_keep=5)
    assert 0 == ShapRFECV._calculate_number_of_features_to_remove(current_num_of_features=5,
                                                                  num_features_to_remove=1,
                                                                  min_num_features_to_keep=5)
    assert 4 == ShapRFECV._calculate_number_of_features_to_remove(current_num_of_features=5,
                                                                  num_features_to_remove=7,
                                                                  min_num_features_to_keep=1)


def test_preprocess_data(X):
    X['col_static'] = 1
    X['col_categorical'] = X['col_3'].apply(lambda x: str(x))
    X['col_missing'] = X['col_3'].apply(lambda x: x if x < 0.5 else np.nan)

    X_output = ShapRFECV._preprocess_data(X)

    # Check if categoricals correctly handled
    assert X_output['col_categorical'].dtype.name == 'category'

    # Check if static feature removed
    assert 'col_static' not in X_output.columns.tolist()
    assert X_output.shape[1] == X.shape[1] - 1


def test_get_feature_shap_values_per_fold(X, y):
    clf = DecisionTreeClassifier(max_depth=1)
    shap_values, train_score, test_score = ShapRFECV._get_feature_shap_values_per_fold(X, y, clf,
                                                                                       train_index=[2, 3, 4, 5, 6, 7],
                                                                                       val_index=[0, 1],
                                                                                       scorer=get_scorer('roc_auc'))
    assert test_score == 1
    assert train_score > 0.9
    assert shap_values.shape == (2, 3)






from probatus.sample_similarity import BaseResemblanceModel, SHAPImportanceResemblance, PermutationImportanceResemblance
from probatus.utils import NotFittedError

import pytest
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.tree import DecisionTreeClassifier


@pytest.fixture(scope='function')
def X1():
    return pd.DataFrame({'col_1': [1, 1, 1, 1],
                         'col_2': [0, 0, 0, 0],
                         'col_3': [0, 0, 0, 0]}, index=[1, 2, 3, 4])


@pytest.fixture(scope='function')
def X2():
    return pd.DataFrame({'col_1': [0, 0, 0, 0],
                         'col_2': [0, 0, 0, 0],
                         'col_3': [0, 0, 0, 0]}, index=[1, 2, 3, 4])


def test_base_class(X1, X2):
    clf = DecisionTreeClassifier(max_depth=1, random_state=1)
    rm = BaseResemblanceModel(clf, test_prc=0.5, n_jobs=1, random_state=42)

    # Before fit it should raise an exception
    with pytest.raises(NotFittedError) as _:
        rm._check_if_fitted()

    actual_report, train_auc, test_auc = rm.fit_compute(X1, X2, return_auc=True)

    # After the fit this should not raise any error
    rm._check_if_fitted()

    assert train_auc == 1
    assert test_auc == 1
    assert actual_report is None

    # Check data splits if correct
    actual_X_train, actual_X_test, actual_y_train, actual_y_test = rm.get_data_splits()

    assert actual_X_train.shape == (4, 3)
    assert actual_X_test.shape == (4, 3)
    assert len(actual_y_train) == 4
    assert len(actual_y_test) == 4

    # Check if stratified
    assert np.sum(actual_y_train) == 2
    assert np.sum(actual_y_test) == 2

    # Check if index is correct
    assert len(rm.X.index.unique()) == 8
    assert list(rm.X.index) == list(rm.y.index)


def test_shap_resemblance_class(X1, X2):
    clf = DecisionTreeClassifier(max_depth=1, random_state=1)
    rm = SHAPImportanceResemblance(clf, test_prc=0.5, n_jobs=1, random_state=42)

    # Before fit it should raise an exception
    with pytest.raises(NotFittedError) as _:
        rm._check_if_fitted()

    actual_report, train_auc, test_auc = rm.fit_compute(X1, X2, return_auc=True)

    # After the fit this should not raise any error
    rm._check_if_fitted()

    assert train_auc == 1
    assert test_auc == 1

    # Check report shape
    assert actual_report.shape == (3, 2)
    # Check if it is sorted by importance
    print(actual_report)
    assert actual_report.iloc[0].name == 'col_1'
    # Check report values
    assert actual_report.loc['col_1']['mean_abs_shap_value'] > 0
    assert actual_report.loc['col_1']['mean_shap_value'] == 0
    assert actual_report.loc['col_2']['mean_abs_shap_value'] == 0
    assert actual_report.loc['col_2']['mean_shap_value'] == 0
    assert actual_report.loc['col_3']['mean_abs_shap_value'] == 0
    assert actual_report.loc['col_3']['mean_shap_value'] == 0

    actual_shap_values_test = rm.get_shap_values()
    assert actual_shap_values_test.shape == (4, 3)

    # Run plots
    rm.plot(plot_type='bar')
    rm.plot(plot_type='dot')


def test_permutation_resemblance_class(X1, X2):
    clf = DecisionTreeClassifier(max_depth=1, random_state=1)
    rm = PermutationImportanceResemblance(clf, test_prc=0.5, n_jobs=1, random_state=42, iterations=20)

    # Before fit it should raise an exception
    with pytest.raises(NotFittedError) as _:
        rm._check_if_fitted()

    actual_report, train_auc, test_auc = rm.fit_compute(X1, X2, return_auc=True)

    # After the fit this should not raise any error
    rm._check_if_fitted()

    assert train_auc == 1
    assert test_auc == 1

    # Check report shape
    assert actual_report.shape == (3, 2)
    # Check if it is sorted by importance
    print(actual_report)
    assert actual_report.iloc[0].name == 'col_1'
    # Check report values
    assert actual_report.loc['col_1']['mean_importance'] > 0
    assert actual_report.loc['col_1']['std_importance'] > 0
    assert actual_report.loc['col_2']['mean_importance'] == 0
    assert actual_report.loc['col_2']['std_importance'] == 0
    assert actual_report.loc['col_3']['mean_importance'] == 0
    assert actual_report.loc['col_3']['std_importance'] == 0

    rm.plot()
    # Check plot size
    fig = plt.gcf()
    size = fig.get_size_inches()
    assert size[0] == 10 and size[1] == 2.5


def test_base_class_same_data(X1):
    clf = DecisionTreeClassifier(max_depth=1, random_state=1)
    rm = BaseResemblanceModel(clf, test_prc=0.5, n_jobs=1, random_state=42)

    actual_report, train_auc, test_auc = rm.fit_compute(X1, X1, return_auc=True)

    assert train_auc == 0.5
    assert test_auc == 0.5
    assert actual_report is None


# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from probatus.binning import SimpleBucketer, AgglomerativeBucketer, QuantileBucketer
from probatus.utils.shap_helpers import shap_to_df
from probatus.utils.arrayfuncs import assure_pandas_df
from probatus.utils.exceptions import NotFittedError


class TreeDependencePlotter:
    """
    Plotter used to plot shap dependence and target rates. 
    
    Args:
        model: classifier for which interpretation is done.

    Example:
    ```python
    from sklearn.datasets import make_classification
    from sklearn.ensemble import RandomForestClassifier

    X, y = make_classification(n_samples=15, n_features=3, n_informative=3, n_redundant=0, random_state=42)
    clf = RandomForestClassifier().fit(X, y)
    bdp = TreeDependencePlotter(clf).fit(X, y)

    bdp.plot(feature=2, type_binning='simple')
    ```
    """

    def __init__(self, model):
        self.model = model

        self.isFitted = False
        self.target_names = [1, 0]

    def __repr__(self):
        return "Shap dependence plotter for {}".format(self.model.__class__.__name__)

    def fit(self, X, y, precalc_shap=None):
        """
        Fits the plotter to the model and data by computing the shap values.
        If the shap_values are passed, they do not need to be computed
        
        Args:
        X (pd.DataFrame): input variables
        y (pd.Series): target variable
        precalc_shap (Optional, None or np.array): Precalculated shap values, If provided they don't need to be
         computed.
        """
        self.X = assure_pandas_df(X)
        self.y = y
        self.features = self.X.columns

        self.shap_vals_df = shap_to_df(self.model, self.X, precalc_shap=precalc_shap)

        self.isFitted = True
        return self


    def _check_fitted(self):
        """
        Function to check if plotter is already fitted, raises exception otherwise
        
        Raises:
            NotFittedError: in case the plotter is not yet fitted.
        """
        if not self.isFitted:
            raise NotFittedError("The plotter is not fitted yet..")

    def plot(
        self,
        feature,
        figsize=(15, 10),
        bins=10,
        type_binning="simple",
        min_q=0,
        max_q=1,
        target_names=None,
    ):
        """
        Plots the shap values for data points for a given feature, as well as the target rate and values distribution.
        
        Args:
            feature (str or int): feature name of the feature to be analyzed.
            figsize ((float, float)): tuple specifying size (width, height) of resulting figure in inches.
            bins (int or list[float]): number of bins or boundaries of bins (supplied in list) for target-rate plot.
            type_binning {'simple', 'agglomerative', 'quantile'}: type of binning to be used in target-rate plot (see :mod:`binning` for more information).
            min_q (float): optional minimum quantile from which to consider values, used for plotting under outliers.
            max_q (float): optional maximum quantile until which data points are considered, used for plotting under outliers.
            target_names (list[str]): optional list of names for target classes to display in plot.
            
        Returns
            matplotlib.pyplot.Figure: Feature plot.
        """
        self._check_fitted()
        if min_q >= max_q:
            raise ValueError("min_q must be smaller than max_q")
        if feature not in self.X.columns:
            raise ValueError("Feature not recognized")
        if type_binning not in ["simple", "agglomerative", "quantile"]:
            raise ValueError(
                "Select one of the following binning methods: 'simple', 'agglomerative', 'quantile'"
            )

        if target_names is not None:
            self.target_names = target_names

        self.min_q, self.max_q = min_q, max_q

        fig = plt.figure(1, figsize=figsize)
        ax1 = plt.subplot2grid((3, 1), (0, 0), rowspan=2)
        ax2 = plt.subplot2grid((3, 1), (2, 0))

        self._dependence_plot(feature=feature, ax=ax1)
        self._target_rate_plot(
            feature=feature, bins=bins, type_binning=type_binning, ax=ax2
        )

        ax2.set_xlim(ax1.get_xlim())

        return fig

    def _dependence_plot(self, feature, ax=None, figsize=(15, 10)):
        """
        Plots shap values for data points with respect to specified feature.
        
        Args:
            feature (str or int): feature for which dependence plot is to be created.
            ax (matplotlib.pyplot.axes): optional axis on which to draw plot.
            figsize ((float, float)): optional tuple with desired figsize in inches.
        
        Returns:
            matplotlib.pyplot.axes: axes on which plot is drawn.
        """
        if type(feature) is int:
            feature = self.features[feature]

        X, y, shap_val = self._get_X_y_shap_with_q_cut(feature=feature)

        ax.scatter(
            X[y == 0], shap_val[y == 0], label=self.target_names[0], color="lightblue"
        )

        ax.scatter(
            X[y == 1], shap_val[y == 1], label=self.target_names[1], color="darkred"
        )

        ax.set_ylabel("Shap value")
        ax.set_title(f"Dependence plot for {feature} feature")
        ax.legend()

        return ax

    def _target_rate_plot(
        self, feature, bins=10, type_binning="simple", ax=None, figsize=(15, 10)
    ):
        """ 
        Plots the distributions of the specific features, as well as the target rate as function of the feature.
        
        Args:
            feature (str or int): feature for which to create target rate plot
            bins (int or list[float]): number of bins or boundaries of desired bins in list.
            type_binning ({'simple', 'agglomerative', 'quantile'}): type of binning strategy used to create bins.
            ax (matplotlib.pyplot.axes): optional axis on which to draw plot.
            figsize ((float, float)): optional tuple with desired figsize in inches.            
            
        Returns:
            bins (list[float]): boundaries of bins used.
            ax (matplotlib.pyplot.axes): axes on which plot is drawn.
            target_ratio (float): total ratio of target (positive over negative).
        """
        x, y, shap_val = self._get_X_y_shap_with_q_cut(feature=feature)

        # Create bins if not explicitly supplied
        if type(bins) is int:
            if type_binning == "simple":
                counts, bins = SimpleBucketer.simple_bins(x, bins)
            elif type_binning == "agglomerative":
                counts, bins = AgglomerativeBucketer.agglomerative_clustering_binning(
                    x, bins
                )
            elif type_binning == "quantile":
                counts, bins = QuantileBucketer.quantile_bins(x, bins)

        # Determine bin for datapoints
        bins[-1] = bins[-1] + 1
        indices = np.digitize(x, bins)

        # Create dataframe with binned data
        dfs = pd.DataFrame(
            {feature: x, "y": y, "bin_index": pd.Series(indices, index=x.index)}
        ).groupby("bin_index", as_index=True)

        # Extract target ratio and mean feature value
        target_ratio = dfs["y"].mean()
        x_vals = dfs[feature].mean()

        # Plot target rate
        ax.hist(x, bins=bins, lw=2, alpha=0.4)
        ax.set_ylabel("Counts")
        ax2 = ax.twinx()
        ax2.plot(x_vals, target_ratio, color="red")
        ax2.set_ylabel("Target rate", color="red", fontsize=12)
        ax2.set_xlim(x.min(), x.max())
        ax.set_xlabel(f'{feature} feature values')

        return bins, ax, target_ratio

    def _get_X_y_shap_with_q_cut(self, feature):
        """
        Extracts all X, y pairs and shap values that fall within defined quantiles of the feature
        
        Args:
            feature (str): feature to return values for
        
        Returns:
            x (pd.Series): selected datapoints
            y (pd.Series): target values of selected datapoints
            shap_val (pd.Series): shap values of selected datapoints
        """
        self._check_fitted()
        if feature not in self.X.columns:
            raise ValueError("Feature not found in data")

        # Prepare arrays
        x = self.X[feature]
        y = self.y
        shap_val = self.shap_vals_df[feature]

        # Determine quantile ranges
        x_min = x.quantile(self.min_q)
        x_max = x.quantile(self.max_q)

        # Create filter
        filter = (x >= x_min) & (x <= x_max)

        # Filter and return terms
        return x[filter], y[filter], shap_val[filter]



# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
from probatus.utils import assure_numpy_array, NotFittedError, get_scorers, warn_if_missing,\
    assure_column_names_consistency
from probatus.utils.shap_helpers import shap_calc, calculate_shap_importance
from sklearn.inspection import permutation_importance
import matplotlib.pyplot as plt
import shap
import warnings

class BaseResemblanceModel(object):
    """
    This model checks for similarity of two samples. A possible use case is analysis whether train sample differs
     from test sample, due to e.g. non-stationarity.

    This is a base class and needs to be extended by a fit() method, which implements how data is split, how model is
     trained and evaluated. Further, inheriting classes need to implement how feature importance should be indicated.
    """
    def __init__(self, model, test_prc=0.25, n_jobs=1, random_state=42):
        """
        Initializes the class.

        Args:
            model (model object): Binary classification model or pipeline.

            test_prc (float, optional): Percentage of data used to test the model. By default 0.25 is set.

            n_jobs (int, optional): Number of parallel executions. If -1 use all available cores. By default 1.

            random_state (int, optional): The seed used by the random number generator.
        """
        self.model = model
        self.test_prc = test_prc
        self.n_jobs = n_jobs
        self.random_state = random_state

        self.fitted = False

        self.metric_name = 'roc_auc'
        self.scorer = get_scorers(self.metric_name)[0]


    def _init_output_variables(self):
        """
        Initializes variables that will be filled in during fit() method, and are used as output
        """
        self.X_train = None
        self.X_test = None
        self.y_train = None
        self.y_test = None
        self.auc_train = None
        self.auc_test = None
        self.report = None


    def fit(self, X1, X2, column_names=None):
        """
        Base fit functionality that should be executed before each fit.

        Args:
            X1 (np.ndarray or pd.DataFrame): First sample to be compared. It needs to have the same number of columns
             as X2.

            X2 (np.ndarray or pd.DataFrame): Second sample to be compared. It needs to have the same number of columns
             as X1.

            column_names (list of str, optional): List of feature names of the provided samples. If provided it will be
             used to overwrite the existing feature names. If not provided the existing feature names are used or default
             feature names are generated.
        """

        # Set seed for results reproducibility
        np.random.seed(self.random_state)

        # Ensure inputs are correct
        self.X1 = assure_numpy_array(X1)
        self.X2 = assure_numpy_array(X2)

        # Check if any missing values
        warn_if_missing(self.X1, 'X1')
        warn_if_missing(self.X2, 'X2')

        # Ensure the same shapes
        if self.X1.shape[1] != self.X2.shape[1]:
            raise(ValueError("Passed variables do not have the same shape. The passed dimensions are {} and {}".
                             format(self.X1.shape[1], self.X2.shape[1])))

        # Check if column_names are passed correctly
        self.column_names = assure_column_names_consistency(column_names, X1)

        # Prepare dataset for modelling
        self.X = pd.DataFrame(np.concatenate([
            self.X1,
            self.X2
        ]), columns = self.column_names).reset_index(drop=True)

        self.y = pd.Series(np.concatenate([
            np.zeros(self.X1.shape[0]),
            np.ones(self.X2.shape[0]),
        ])).reset_index(drop=True)

        # Reinitialize variables in case of multiple times being fit
        self._init_output_variables()

        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(self.X, self.y,test_size=self.test_prc,
                                                                                random_state=self.random_state,
                                                                                stratify=self.y)
        self.model.fit(self.X_train, self.y_train)

        self.auc_train = np.round(self.scorer.score(self.model, self.X_train, self.y_train), 3)
        self.auc_test = np.round(self.scorer.score(self.model, self.X_test, self.y_test), 3)


        print(f'Finished model training: Train AUC {self.auc_train},'
              f' Test AUC {self.auc_test}')

        if self.auc_train > self.auc_test:
            warnings.warn('Train AUC > Test AUC, which might indicate an overfit. \n'
                          'Strong overfit might lead to misleading conclusions when analysing feature importance. '
                          'Consider retraining with more regularization applied to the model.')

        self.fitted = True


    def _check_if_fitted(self):
        if self.fitted is False:
            raise(NotFittedError('The object has not been fitted. Please run fit() method first'))


    def get_data_splits(self):
        """
        Returns the data splits used to train the Resemblance model.

        Returns:
            (pd.DataFrame, pd.DataFrame, pd.Series, pd.Series): X_train, X_test, y_train, y_test.
        """
        self._check_if_fitted()
        return self.X_train, self.X_test, self.y_train, self.y_test


    def compute(self, return_auc=False):
        """
        Checks if fit() method has been run and computes the output variables.

        Args:
            return_auc (bool, optional): Flag indicating whether the method should return a tuple (feature
             importances, train AUC, test AUC), or feature importances. By default the second option is selected.

        Returns:
            tuple(pd.DataFrame, float, float) or pd.DataFrame: Depending on value of return_tuple either returns a
            tuple (feature importances, train AUC, test AUC), or feature importances.
        """
        self._check_if_fitted()

        if return_auc:
            return self.report, self.auc_train, self.auc_test
        else:
            return self.report


    def fit_compute(self, X1, X2, column_names=None, return_auc=False, **fit_kwargs):
        """
        Fits the resemblance model and computes the report regarding feature importance.

        Args:
            X1 (np.ndarray or pd.DataFrame): First sample to be compared. It needs to have the same number of columns
             as X2.

            X2 (np.ndarray or pd.DataFrame): Second sample to be compared. It needs to have the same number of columns
             as X1.

            column_names (list of str, optional): List of feature names of the provided samples. If provided it will be
             used to overwrite the existing feature names. If not provided the existing feature names are used or default
             feature names are generated.

            return_auc (bool, optional): Flag indicating whether the method should return a tuple (feature
             importances, train AUC, test AUC), or only feature importances. By default the second option is selected.

            **fit_kwargs: keyword arguments passed to the fit() method:

                - `X1`: First sample to be compared. It needs to have the same number of columns as X2.
                - `X2`: Second sample to be compared. It needs to have the same number of columns as X1.
                - `column_names`: List of feature names of the provided samples. If provided it will be
                 used to overwrite the existing feature names. If not provided the existing feature names are used or default
                 feature names are generated.

        Returns:
            tuple of (pd.DataFrame, float, float) or pd.DataFrame: Depending on value of return_tuple either returns a
             tuple (feature importances, train AUC, test AUC), or feature importances.
        """
        self.fit(X1, X2, column_names=column_names, **fit_kwargs)
        return self.compute(return_auc=return_auc)


class PermutationImportanceResemblance(BaseResemblanceModel):
    """
    This model checks for similarity of two samples. A possible use case is analysis whether train sample differs
     from test sample, due to e.g. non-stationarity.

    It assigns to labels to each sample, 0 to first sample, 1 to the second. Then, It randomly selects a portion of
     data to train on. The resulting model tries to distinguish which sample does a given test row comes from. This
     provides insights on how distinguishable these samples are and which features contribute to that. The feature
     importance is calculated using permutation importance.

    If the model achieves test AUC significantly different than 0.5, it indicates that it is possible to distinguish
     the samples, and therefore, the samples differ. Features with high permutation importance contribute to that
     effect the most. Thus, their distribution might differ between two samples.

    Examples:
    ```python
    from sklearn.datasets import make_classification
    from sklearn.ensemble import RandomForestClassifier
    from probatus.sample_similarity import PermutationImportanceResemblance
    X1, _ = make_classification(n_samples=1000, n_features=5)
    X2, _ = make_classification(n_samples=1000, n_features=5, shift=0.5)
    clf = RandomForestClassifier(max_depth=2)
    perm = PermutationResemblanceModel(clf)
    feature_importance = perm.fit_compute(X1, X2)
    perm.plot()
    ```
    """

    def __init__(self, model, iterations=100, **kwargs):
        """
        Initializes the class.

        Args:
            model (model object): Binary classification model or pipeline.

            iterations (int, optional): Number of iterations performed to calculate permutation importance. By default 100
             iterations per feature are done.

            **kwargs: Keyword arguments that can overwrite inherited default values in BaseResemblanceModel:

                - `test_prc`: Percentage of data used to test the model. By default 0.25 is set.
                - `n_jobs`: Number of parallel executions. If -1 use all available cores. By default 1.
                - `random_state`: The seed used by the random number generator.
        """
        super().__init__(model=model, **kwargs)

        self.iterations = iterations

        self.iterations_columns = ['feature', 'importance']
        self.iterations_results = pd.DataFrame(columns=self.iterations_columns)

        self.plot_x_label = 'Permutation Feature Importance'
        self.plot_y_label = 'Feature Name'
        self.plot_title = 'Permutation Feature Importance of Resemblance Model'


    def fit(self, X1, X2, column_names=None):
        """
        This function assigns to labels to each sample, 0 to first sample, 1 to the second. Then, It randomly selects a
        portion of data to train on. The resulting model tries to distinguish which sample does a given test row comes
        from. This provides insights on how distinguishable these samples are and which features contribute to that. The
        feature importance is calculated using permutation importance.

        Args:
            X1 (np.ndarray or pd.DataFrame): First sample to be compared. It needs to have the same number of columns
             as X2.

            X2 (np.ndarray or pd.DataFrame): Second sample to be compared. It needs to have the same number of columns
             as X1.

            column_names (list of str, optional): List of feature names of the provided samples. If provided it will be
             used to overwrite the existing feature names. If not provided the existing feature names are used or default
             feature names are generated.
        """
        super().fit(X1=X1, X2=X2, column_names=column_names)


        permutation_result = permutation_importance(self.model, self.X_test, self.y_test, scoring=self.scorer.scorer,
                                                    n_repeats=self.iterations, n_jobs=self.n_jobs)

        # Prepare report
        self.report_columns = ['mean_importance', 'std_importance']
        self.report = pd.DataFrame(index=self.column_names, columns=self.report_columns, dtype=float)

        for feature_index, feature_name in enumerate(self.column_names):
            # Fill in the report
            self.report.loc[feature_name, 'mean_importance'] =\
                permutation_result['importances_mean'][feature_index]
            self.report.loc[feature_name, 'std_importance'] =\
                permutation_result['importances_std'][feature_index]

            # Fill in the iterations
            current_iterations = pd.DataFrame(
                np.stack([
                    np.repeat(feature_name, self.iterations),
                    permutation_result['importances'][feature_index, :].reshape((self.iterations,))
                    ], axis=1),
                columns=self.iterations_columns)

            self.iterations_results = pd.concat([self.iterations_results, current_iterations])

        self.iterations_results['importance'] = self.iterations_results['importance'].astype(float)

        # Sort by mean test score of first metric
        self.report.sort_values(by='mean_importance', ascending=False, inplace=True)


    def plot(self, ax=None, top_n=None):
        """
        Plots the resulting AUC of the model as well as the feature importances.

        Args:
            ax (matplotlib.axes, optional): Axes to which the output should be plotted. If not provided a new axes are
             created.

            top_n (int, optional): Number of the most important features to be plotted. By default are features are
             included into the plot.

        Returns:
            matplotlib.axes, optional: Axes that include the plot.
        """

        feature_report =  self.compute()
        self.iterations_results['importance'] =  self.iterations_results['importance'].astype(float)

        sorted_features = feature_report['mean_importance'].\
            sort_values(ascending=True).index.values
        if top_n is not None and top_n > 0:
            sorted_features = sorted_features[-top_n:]

        if ax is None:
            height_per_subplot = len(sorted_features) / 2. + 1
            width_per_subplot = 10
            fig, ax = plt.subplots(figsize=(width_per_subplot, height_per_subplot))

        for position, feature in enumerate(sorted_features):
            ax.boxplot(self.iterations_results[self.iterations_results['feature']==feature]['importance'],
                       positions=[position], vert=False)

        ax.set_yticks(range(position + 1))
        ax.set_yticklabels(sorted_features)
        ax.set_xlabel(self.plot_x_label)
        ax.set_ylabel(self.plot_y_label)
        ax.set_title(self.plot_title)

        fig_text = "Train AUC: {},\n" \
                   "Test AUC: {}.". \
                       format(self.auc_train, self.auc_test)

        ax.annotate(fig_text, (0,0), (0, -50), fontsize=12, xycoords='axes fraction',
                    textcoords='offset points', va='top')

        return ax


class SHAPImportanceResemblance(BaseResemblanceModel):
    """
    This model checks for similarity of two samples. A possible use case is analysis whether train sample differs
     from test sample, due to e.g. non-stationarity.

    It assigns to labels to each sample, 0 to first sample, 1 to the second. Then, It randomly selects a portion of
     data to train on. The resulting model tries to distinguish which sample does a given test row comes from. This
     provides insights on how distinguishable these samples are and which features contribute to that. The feature
     importance is calculated using SHAP feature importance.

    If the model achieves test AUC significantly different than 0.5, it indicates that it is possible to distinguish
     the samples, and therefore, the samples differ. Features with high permutation importance contribute to that
     effect the most. Thus, their distribution might differ between two samples.

    This class currently works only with the Tree based models.

    Examples:
    ```python
    from sklearn.datasets import make_classification
    from sklearn.ensemble import RandomForestClassifier
    from probatus.sample_similarity import SHAPImportanceResemblance
    X1, _ = make_classification(n_samples=1000, n_features=5)
    X2, _ = make_classification(n_samples=1000, n_features=5, shift=0.5)
    clf = RandomForestClassifier(max_depth=2)
    rm = SHAPImportanceResemblance(clf)
    feature_importance = rm.fit_compute(X1, X2)
    rm.plot()
    ```
    """

    def __init__(self, model,  **kwargs):
        """
        Initializes the class.

        Args:
            model (model object): Binary classification model or pipeline.

            **kwargs: Keyword arguments that can overwrite inherited default values in BaseResemblanceModel:

                - `test_prc`: Percentage of data used to test the model. By default 0.25 is set.
                - `n_jobs`: Number of parallel executions. If -1 use all available cores. By default 1.
                - `random_state`: The seed used by the random number generator.
        """
        super().__init__(model=model, **kwargs)

        self.plot_title = 'SHAP summary plot'

    def fit(self, X1, X2, column_names=None):
        """
        This function assigns to labels to each sample, 0 to first sample, 1 to the second. Then, It randomly selects a
         portion of data to train on. The resulting model tries to distinguish which sample does a given test row comes
         from. This provides insights on how distinguishable these samples are and which features contribute to that. The
         feature importance is calculated using SHAP feature importance.

        Args:
            X1 (np.ndarray or pd.DataFrame): First sample to be compared. It needs to have the same number of columns
             as X2.

            X2 (np.ndarray or pd.DataFrame): Second sample to be compared. It needs to have the same number of columns
             as X1.

            column_names (list of str, optional): List of feature names of the provided samples. If provided it will be
             used to overwrite the existing feature names. If not provided the existing feature names are used or default
             feature names are generated.
        """
        super().fit(X1=X1, X2=X2, column_names=column_names)

        self.shap_values_test = shap_calc(self.model, self.X_test, data=self.X_train)
        self.report = calculate_shap_importance(self.shap_values_test, self.column_names)

    def plot(self, plot_type='bar', **summary_plot_kwargs):
        """
        Plots the resulting AUC of the model as well as the feature importances.

        Args:
            plot_type (Optional, str): Type of plot, used to compute shap.summary_plot. By default 'bar', available ones
             are  "dot", "bar", "violin",

            **summary_plot_kwargs: kwargs passed to the shap.summary_plot.

        Returns:
            matplotlib.axes, optional: Axes that include the plot.
        """

        # This line serves as a double check if the object has been fitted
        self._check_if_fitted()

        shap.summary_plot(self.shap_values_test, self.X_test, plot_type=plot_type,
                          class_names=['First Sample', 'Second Sample'], show=False, **summary_plot_kwargs)
        ax = plt.gca()
        ax.set_title(self.plot_title)

        fig_text = "Train AUC: {},\n" \
                   "Test AUC: {}.". \
                       format(self.auc_train, self.auc_test)

        ax.annotate(fig_text, (0,0), (0, -50), fontsize=12, xycoords='axes fraction',
                    textcoords='offset points', va='top')
        return ax


    def get_shap_values(self):
        '''
        Gets the SHAP values generated on the test set.

        Returns:
             (np.array) SHAP values generated on the test set.
        '''
        self._check_if_fitted()
        return self.shap_values_test




from sklearn.tree import DecisionTreeClassifier
import pytest
import numpy as np
import pandas as pd
from probatus.interpret import ShapModelInterpreter
from unittest.mock import patch

@pytest.fixture(scope='function')
def X_train():
    return pd.DataFrame({'col_1': [1, 1, 1, 1],
                         'col_2': [0, 0, 0, 0],
                         'col_3': [1, 0, 1, 0]}, index=[1, 2, 3, 4])


@pytest.fixture(scope='function')
def y_train():
    return pd.Series([1, 0, 1, 0], index=[1, 2, 3, 4])


@pytest.fixture(scope='function')
def X_test():
    return pd.DataFrame({'col_1': [1, 1, 1, 1],
                         'col_2': [0, 0, 0, 0],
                         'col_3': [1, 0, 1, 0]}, index=[5, 6, 7, 8])


@pytest.fixture(scope='function')
def y_test():
    return pd.Series([0, 0, 1, 0], index=[5, 6, 7, 8])


@pytest.fixture(scope='function')
def fitted_tree(X_train, y_train):
    return DecisionTreeClassifier(max_depth=1, random_state=1).fit(X_train, y_train)


@pytest.fixture(scope='function')
def expected_feature_importance():
    return pd.DataFrame({
        'mean_abs_shap_value': [0.5, 0., 0.],
        'mean_shap_value': [0., 0., 0.]}, index=['col_3', 'col_1', 'col_2'])


def test_shap_interpret(fitted_tree, X_train, y_train, X_test, y_test, expected_feature_importance):
    class_names = ['neg', 'pos']

    shap_interpret = ShapModelInterpreter(fitted_tree)
    shap_interpret.fit(X_train, X_test, y_train, y_test, class_names=class_names)

    # Check parameters
    assert shap_interpret.fitted == True
    shap_interpret._check_if_fitted

    assert shap_interpret.class_names == class_names
    assert shap_interpret.auc_train == 1
    assert shap_interpret.auc_test == pytest.approx(0.833, 0.01)

    # Check expected shap values
    assert (np.mean(np.abs(shap_interpret.shap_values), axis=0) == [0, 0, 0.5]).all()

    importance_df = shap_interpret.compute()
    pd.testing.assert_frame_equal(expected_feature_importance, importance_df)

    with patch('matplotlib.pyplot.figure') as mock_plt:
        with patch('shap.plots._waterfall.waterfall_legacy'):
            ax1 = shap_interpret.plot('importance')
            ax2 = shap_interpret.plot('summary')
            ax3 =shap_interpret.plot('dependence', target_columns='col_3')
            ax4 = shap_interpret.plot('sample', samples_index=[5, 6])
    assert not(isinstance(ax1, list))
    assert not(isinstance(ax2, list))
    assert not(isinstance(ax3, list))
    assert isinstance(ax4, list) and len(ax4) == 2


def test_shap_interpret_fit_compute(fitted_tree, X_train, y_train, X_test, y_test, expected_feature_importance):
    class_names = ['neg', 'pos']

    shap_interpret = ShapModelInterpreter(fitted_tree)
    importance_df = shap_interpret.fit_compute(X_train, X_test, y_train, y_test, class_names=class_names)

    # Check parameters
    assert shap_interpret.fitted == True
    shap_interpret._check_if_fitted

    assert shap_interpret.class_names == class_names
    assert shap_interpret.auc_train == 1
    assert shap_interpret.auc_test == pytest.approx(0.833, 0.01)

    # Check expected shap values
    assert (np.mean(np.abs(shap_interpret.shap_values), axis=0) == [0, 0, 0.5]).all()

    pd.testing.assert_frame_equal(expected_feature_importance, importance_df)


# Copyright (c) 2020 ING Bank N.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import numpy as np
import pandas as pd
import numbers

from probatus.utils import DimensionalityError
import warnings


def check_1d(x):
    """
    Checks whether or not a list, numpy array, pandas dataframe, pandas series are one-dimensional.
    Returns True when check is ok, otherwise throws a `DimensionalityError`

    Args:
        x: list, numpy array, pandas dataframe, pandas series

    Returns: True or throws `DimensionalityError`

    """
    if isinstance(x, list):
        if any([isinstance(el, list) for el in x]):
            raise DimensionalityError('The input is not 1D')
        else:
            return True
    if isinstance(x, np.ndarray):
        if x.ndim == 1 and all([isinstance(el, numbers.Number) for el in x]):
            return True
        else:
            raise DimensionalityError('The input is not 1D')
    if isinstance(x, pd.core.frame.DataFrame):
        if len(x.columns) == 1 and pd.api.types.is_numeric_dtype(
                x[x.columns[0]]):
            return True
        else:
            raise DimensionalityError('The input is not 1D')
    if isinstance(x, pd.core.series.Series):
        if x.ndim == 1 and pd.api.types.is_numeric_dtype(x):
            return True
        else:
            raise DimensionalityError('The input is not 1D')


def assure_numpy_array(x, assure_1d=False):
    """
    Returns x as numpy array. X can be a list, numpy array, pandas dataframe, pandas series

    Args:
        x: list, numpy array, pandas dataframe, pandas series
        assure_1d: whether or not to assure that the input x is one-dimensional

    Returns: numpy array

    """
    if assure_1d:
        _ = check_1d(x)
    if isinstance(x, list):
        return np.array(x)
    if isinstance(x, np.ndarray):
        return x
    if isinstance(x, pd.core.frame.DataFrame):
        if len(x.columns) == 1:
            return x.values.flatten()
        else:
            return x.values
    if isinstance(x, pd.core.series.Series):
        return x.values


def assure_pandas_df(x):
    """
    Returns x as pandas DataFrame. X can be a list, list of lists, numpy array, pandas DataFrame or pandas Series
    
    Args:
        x (list, numpy array, pandas DataFrame, pandas Series): array to be tested
        
    Returns:
        pandas DataFrame
    """
    if isinstance(x, pd.core.frame.DataFrame):
        return x
    elif any([
            isinstance(x, np.ndarray),
            isinstance(x, pd.core.series.Series),
            isinstance(x, list)
    ]):
        return pd.DataFrame(x)
    else:
        raise TypeError(
            "Please supply a list, numpy array, pandas Series or pandas DataFrame"
        )


def assure_pandas_series(x, index=None):
    """
    Returns x as pandas Series. X can be a list, numpy array, or pandas Series

    Args:
        x (list, numpy array, pandas DataFrame, pandas Series): array to be tested

    Returns:
        pandas Series
    """
    if isinstance(x, pd.Series):
        return x
    elif any([
        isinstance(x, np.ndarray),
        isinstance(x, list)
    ]):
        return pd.Series(x, index=index)
    else:
        raise TypeError(
            "Please supply a list, numpy array, pandas Series"
        )

def assure_column_names_consistency(column_names, df):
    """
    Ensure that the column names are correct. If they are None, then, the column names from df are taken. Otherwise,
    the function checks if the passed list has the correct shape, to be consistent with the size of the df.

    Args:
        column_names (None, or list of str): List of column names to use for the df.
        df (pd.DataFrame): Dataset.

    Returns:
        (list of str): List of column names.
    """
    # Check if column_names are passed correctly
    if column_names is None:
        # Checking if original X1 was a df then taking its column names
        if isinstance(df, pd.DataFrame):
            column_names = df.columns.tolist()
        # Otherwise make own feature names
        else:
            column_names = ['column_{}'.format(idx) for idx in range(df.shape[1])]
    else:
        if isinstance(column_names, list):
            if len(column_names) == df.shape[1]:
                column_names = column_names
            else:
                raise (ValueError("Passed column_names have different dimensionality than input samples. "
                                  "The dimensionality of column_names is {} and first sample {}".
                                  format(len(column_names), df.shape[1])))
        else:
            raise (TypeError("Passed column_names must be a list"))
    return column_names


def warn_if_missing(variable, variable_name):
    """
    Checks if for missing values: if there are notify the user
    Args:
        variable (pandas.DataFrame, pandas.Series or numpy.ndarray): data to be checked for missing values.
        variable_name (str): Name of the variable checked.
    """
    warning_text = "You have missing values in your variable {}, this might cause the model to fail. Please either " \
                   "impute the missing values or use a model that can handle them e.g. XGBoost.".format(variable_name)

    if isinstance(variable, (pd.DataFrame, pd.Series)):
        if variable.isnull().values.any():
            warnings.warn(warning_text)
    if isinstance(variable, np.ndarray):
        if pd.isnull(variable).any():
            warnings.warn(warning_text)


def check_numeric_dtypes(x):
    """
    Checks if all entries in an array are of a data type that can be interpreted as numeric (int, float or bool)
    
    Args:
        x (np.ndarray or pd.Series, list): array to be checked
    
    Returns:
        x: unchanged input array
        
    Raises:
        TypeError: if not all elements are of numeric dtypes
    """
    x = assure_numpy_array(x)
    allowed_types = [bool, int, float]
    
    for element in np.nditer(x):
        if type(element.item()) not in allowed_types:
            raise TypeError("Please supply an array with only floats, ints or booleans")
    return x


